@@ -752,6 +752,22 @@ impl<F: File<Node>, D: Dir, T: Transaction<D>> Restore<D, T> for BTreeFile<F, D,
     }
 }
 
+#[async_trait]
+impl<F: File<Node>, D: Dir, T: Transaction<D>, I: BTreeInstance> CopyFrom<D, T, I>
+    for BTreeFile<F, D, T>
+{
+    async fn copy_from(source: I, file: F, txn_id: TxnId) -> TCResult<Self>
+    where
+        I: 'async_trait,
+    {
+        let schema = source.schema().clone();
+        let dest = Self::create(file, schema, txn_id).await?;
+        let keys = source.keys(txn_id).await?;
+        dest.try_insert_from(txn_id, keys).await?;
+        Ok(dest)
+    }
+}
+
 impl<F, D, T> fmt::Display for BTreeFile<F, D, T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         f.write_str("a BTree")
