@@ -8,14 +8,17 @@ use async_trait::async_trait;
 use destream::{de, en};
 use futures::{future, Stream, TryFutureExt, TryStreamExt};
 use log::debug;
-use safecast::{Match, TryCastFrom, TryCastInto};
+use safecast::*;
 
 use tc_error::*;
-use tc_transact::fs::{Dir, File};
+use tc_transact::fs::{Dir, File, Hash};
 use tc_transact::{IntoView, Transaction, TxnId};
 use tc_value::{Link, NumberType, Value, ValueCollator, ValueType};
 use tcgeneric::*;
 
+pub use file::{BTreeFile, Node};
+pub use slice::BTreeSlice;
+
 mod file;
 mod slice;
 
@@ -24,9 +27,6 @@ const PREFIX: PathLabel = path_label(&["state", "collection", "btree"]);
 /// The file extension of a [`BTree`] as stored on-disk
 pub const EXT: &str = "node";
 
-pub use file::{BTreeFile, Node};
-pub use slice::BTreeSlice;
-
 /// A [`BTree`] key.
 pub type Key = Vec<Value>;
 
@@ -77,13 +77,21 @@ pub trait BTreeInstance: Clone + Instance {
     async fn try_insert_from<S: Stream<Item = TCResult<Key>> + Send + Unpin>(
         &self,
         txn_id: TxnId,
-        mut keys: S,
+        keys: S,
     ) -> TCResult<()> {
-        while let Some(key) = keys.try_next().await? {
-            self.insert(txn_id, key).await?;
-        }
+        keys.map_ok(|key| self.insert(txn_id, key))
+            .try_buffer_unordered(num_cpus::get())
+            .try_fold((), |(), ()| future::ready(Ok(())))
+            .await
+    }
+}
 
-        Ok(())
+#[async_trait]
+impl<'en, F: File<Node>, D: Dir, T: Transaction<D>> Hash<'en> for BTree<F, D, T> {
+    type Item = Key;
+
+    async fn hashable(&'en self, txn_id: TxnId) -> TCResult<TCTryStream<'en, Self::Item>> {
+        self.clone().keys(txn_id).await
     }
 }
 
@@ -186,6 +194,19 @@ impl TryCastFrom<Value> for Column {
     }
 }
 
+impl From<Column> for Value {
+    fn from(column: Column) -> Self {
+        Value::Tuple(
+            vec![
+                column.name.into(),
+                column.dtype.path().into(),
+                column.max_len.map(Value::from).into(),
+            ]
+            .into(),
+        )
+    }
+}
+
 struct ColumnVisitor;
 
 #[async_trait]
